#include <Arduino.h>
#include <SPI.h>
#include <SD.h>
#include "Ucglib.h"
#include "AxisJoystick.h"

#define LED_STATS A0
#define LED_ITEMS A1
#define LED_DATA A2

#define VRX A3
#define VRY A4
#define SW 5

#define LCD_SCK 13
#define LCD_MISO 12
#define LCD_MOSI 11
#define LCD_CS 10
#define LCD_DC 9
#define LCD_RES 8

#define BUZZER 7
#define SD_CS 4
#define BUTTON 2

#define BUFFPIXEL 20

Ucglib_ILI9341_18x240x320_HWSPI ucg(LCD_DC, LCD_CS, LCD_RES);

Joystick* joystick;
int joystick_select = 0;
int window = 0;
int show_window = 0;
long ts = 0;
long j_ts = 0;

void setup()
{
  Serial.begin(9600);
  
  pinMode(BUTTON, INPUT_PULLUP);
  attachInterrupt(digitalPinToInterrupt(BUTTON), pin_ISR, FALLING);

  joystick = new AxisJoystick(SW, VRX, VRY);
  joystick->calibrate(0, 1023, 100); // low range, high range, range division
  
  pinMode(BUZZER, OUTPUT);

  pinMode(LED_STATS, OUTPUT);
  pinMode(LED_ITEMS, OUTPUT);
  pinMode(LED_DATA, OUTPUT);
 
  setup_sd();  
  
  ucg.begin(UCG_FONT_MODE_TRANSPARENT);
  setup_ucg();
  show_stats();
  
  digitalWrite(LED_STATS, HIGH);
  digitalWrite(LED_ITEMS, LOW);
  digitalWrite(LED_DATA, LOW);
}

void loop() {
  
  if (show_window) {
    show_window = 0;
    
    setup_ucg();
    switch(window) {
      case 1: show_items(); break;
      case 2: show_data(); break;
      default: show_stats();
    }

    make_sound();

  }

    read_joystick();
}

void pin_ISR() {
  
   if (millis() - ts > 300) {
      ts = millis();
      show_window = 1;
      switch(window) {
      case 1: 
              window = 2; //show data
              break;
      case 2:
              window = 0; // show stats
              break;
      default: 
              window = 1; // show items
    }
  }
}

void show_stats() {
  //ucg.print("Stats");
  char fname[] = "stats.bmp";
  bmpDraw(fname, 0, 0);

  digitalWrite(LED_DATA, LOW);
  digitalWrite(LED_STATS, HIGH);
}

void show_items() {
 // ucg.print("Items");
 char fname[] = "items.bmp";
  bmpDraw(fname, 0, 0);
 
 digitalWrite(LED_STATS, LOW);
 digitalWrite(LED_ITEMS, HIGH);
}

void show_data() {
//  ucg.print("Data");
 char fname[] = "data.bmp";
  bmpDraw(fname, 0, 0);
 
  digitalWrite(LED_ITEMS, LOW);
  digitalWrite(LED_DATA, HIGH);   
}

void make_sound() {
  tone(BUZZER, 659, 200);
  delay(200);
}

void read_joystick() {
   if (millis() - j_ts > 300 &&
      (joystick->isUp() || joystick->isDown() || joystick->isLeft() || joystick->isRight())) {
      j_ts = millis();
      Serial.print(" | Up: " + String(joystick->isUp()));
      Serial.print(" | Down: " + String(joystick->isDown()));
      Serial.print(" | Right: " + String(joystick->isRight()));
      Serial.println(" | Left: " + String(joystick->isLeft()));
    }
}

void setup_ucg() {
  ucg.clearScreen();
  
  ucg.setFont(ucg_font_ncenR12_tr);
  ucg.setColor(255, 255, 255);
  ucg.setColor(1, 255, 0,0);
  
  ucg.setPrintPos(0,25);
}

void setup_sd() {
   while (!Serial) {
    ; // wait for serial port to connect. Needed for native USB port only
  }

  Serial.print("Initializing SD card...");

  if (!SD.begin(SD_CS)) {
    Serial.println("initialization failed!");
    while (1);
  }
  Serial.println("initialization done.");
}

void bmpDraw(char *filename, int x, int y) {

  File     bmpFile;
  int      bmpWidth, bmpHeight;   // W+H in pixels
  uint8_t  bmpDepth;              // Bit depth (currently must be 24)
  uint32_t bmpImageoffset;        // Start of image data in file
  uint32_t rowSize;               // Not always = bmpWidth; may have padding
  uint8_t  sdbuffer[3*BUFFPIXEL]; // pixel in buffer (R+G+B per pixel)
  uint8_t  buffidx = sizeof(sdbuffer); // Current position in sdbuffer
  boolean  goodBmp = false;       // Set to true on valid header parse
  boolean  flip    = true;        // BMP is stored bottom-to-top
  int      w, h, row, col;
  uint8_t  r, g, b;
  uint32_t pos = 0, startTime = millis();
  boolean  first = true;

  if((x >= ucg.getWidth()) || (y >= ucg.getHeight())) return;

  Serial.println();
  progmemPrint(PSTR("Loading image '"));
  Serial.print(filename);
  Serial.println('\'');
  
  // Open requested file on SD card
  if (!(bmpFile = SD.open(filename, FILE_READ))) {
    progmemPrintln(PSTR("File not found"));
    return;
  }

  
  // Parse BMP header
  if(read16(bmpFile) == 0x4D42) { // BMP signature
    progmemPrint(PSTR("File size: ")); Serial.println(read32(bmpFile));
    (void)read32(bmpFile); // Read & ignore creator bytes
    bmpImageoffset = read32(bmpFile); // Start of image data
    progmemPrint(PSTR("Image Offset: ")); Serial.println(bmpImageoffset, DEC);
    // Read DIB header
    progmemPrint(PSTR("Header size: ")); Serial.println(read32(bmpFile));
    bmpWidth  = read32(bmpFile);
    bmpHeight = read32(bmpFile);
    if(read16(bmpFile) == 1) { // # planes -- must be '1'
      bmpDepth = read16(bmpFile); // bits per pixel
      progmemPrint(PSTR("Bit Depth: ")); Serial.println(bmpDepth);
      if((bmpDepth == 24) && (read32(bmpFile) == 0)) { // 0 = uncompressed

        goodBmp = true; // Supported BMP format -- proceed!
        progmemPrint(PSTR("Image size: "));
        Serial.print(bmpWidth);
        Serial.print('x');
        Serial.println(bmpHeight);
        
        // BMP rows are padded (if needed) to 4-byte boundary
        rowSize = (bmpWidth * 3 + 3) & ~3;

        // Crop area to be loaded
        w = bmpWidth;
        h = bmpHeight;
        
        for (row = 0; row < h; row++) { // For each scanline...
          // Seek to start of scan line.  It might seem labor-
          // intensive to be doing this on every line, but this
          // method covers a lot of gritty details like cropping
          // and scanline padding.  Also, the seek only takes
          // place if the file position actually needs to change
          // (avoids a lot of cluster math in SD library).
            pos = bmpImageoffset + row * rowSize;
          if(bmpFile.position() != pos) { // Need seek?
            bmpFile.seek(pos);
            buffidx = sizeof(sdbuffer); // Force buffer reload
          }

          for (col = 0; col < w; col++) { // For each column...
            // Time to read more pixel data?
            if (buffidx >= sizeof(sdbuffer)) { // Indeed
              bmpFile.read(sdbuffer, sizeof(sdbuffer));
              buffidx = 0; // Set index to beginning
            }

            // Convert pixel from BMP to TFT format
            b = sdbuffer[buffidx++];
            g = sdbuffer[buffidx++];
            r = sdbuffer[buffidx++];
            
            ucg.setColor(r,g,b);
            ucg.drawPixel(row, col);
          } // end pixel
        } // end scanline
        
        Serial.println("Image loaded"); 
      } // end goodBmp
    }
  }

  bmpFile.close();
  if(!goodBmp) progmemPrintln(PSTR("BMP format not recognized."));
}

// These read 16- and 32-bit types from the SD card file.
// BMP data is stored little-endian, Arduino is little-endian too.
// May need to reverse subscript order if porting elsewhere.

uint16_t read16(File f) {
  uint16_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read(); // MSB
  return result;
}

uint32_t read32(File f) {
  uint32_t result;
  ((uint8_t *)&result)[0] = f.read(); // LSB
  ((uint8_t *)&result)[1] = f.read();
  ((uint8_t *)&result)[2] = f.read();
  ((uint8_t *)&result)[3] = f.read(); // MSB
  return result;
}

// Copy string from flash to serial port
// Source string MUST be inside a PSTR() declaration!
void progmemPrint(const char *str) {
  char c;
  while(c = pgm_read_byte(str++)) Serial.print(c);
}

// Same as above, with trailing newline
void progmemPrintln(const char *str) {
  progmemPrint(str);
  Serial.println();
}
